@@ -1,4 +1,3 @@
-import ConfigParser
 import socket
 import getopt
 import os
@@ -7,7 +6,6 @@ import tempfile
 import errno
 import signal
 import re
-import xmlrpclib
 import pwd
 import grp
 import resource
@@ -18,6 +16,13 @@ import platform
 import warnings
 import fcntl
 
+from supervisor.compat import PY2
+from supervisor.compat import ConfigParser
+from supervisor.compat import as_bytes, as_string
+from supervisor.compat import xmlrpclib
+from supervisor.compat import StringIO
+from supervisor.compat import basestring
+
 from supervisor.medusa import asyncore_25 as asyncore
 
 from supervisor.datatypes import process_or_group_name
@@ -49,9 +54,12 @@ from supervisor import states
 from supervisor import xmlrpc
 from supervisor import poller
 
-mydir = os.path.abspath(os.path.dirname(__file__))
-version_txt = os.path.join(mydir, 'version.txt')
-VERSION = open(version_txt).read().strip()
+def _read_version_txt():
+    mydir = os.path.abspath(os.path.dirname(__file__))
+    version_txt = os.path.join(mydir, 'version.txt')
+    with open(version_txt, 'r') as f:
+        return f.read().strip()
+VERSION = _read_version_txt()
 
 def normalize_path(v):
     return os.path.normpath(os.path.abspath(os.path.expanduser(v)))
@@ -194,7 +202,7 @@ class Options:
             if rest not in ("", ":"):
                 raise ValueError("short option should be 'x' or 'x:'")
             key = "-" + key
-            if self.options_map.has_key(key):
+            if key in self.options_map:
                 raise ValueError("duplicate short option key '%s'" % key)
             self.options_map[key] = (name, handler)
             self.short_options.append(short)
@@ -206,7 +214,7 @@ class Options:
             if key[-1] == "=":
                 key = key[:-1]
             key = "--" + key
-            if self.options_map.has_key(key):
+            if key in self.options_map:
                 raise ValueError("duplicate long option key '%s'" % key)
             self.options_map[key] = (name, handler)
             self.long_options.append(long)
@@ -245,8 +253,11 @@ class Options:
         if progname is None:
             progname = sys.argv[0]
         if doc is None:
-            import __main__
-            doc = __main__.__doc__
+            try:
+                import __main__
+                doc = __main__.__doc__
+            except Exception:
+                pass
         self.progname = progname
         self.doc = doc
 
@@ -257,7 +268,7 @@ class Options:
         try:
             self.options, self.args = getopt.getopt(
                 args, "".join(self.short_options), self.long_options)
-        except getopt.error, exc:
+        except getopt.error as exc:
             self.usage(str(exc))
 
         # Check for positional args
@@ -270,7 +281,7 @@ class Options:
             if handler is not None:
                 try:
                     arg = handler(arg)
-                except ValueError, msg:
+                except ValueError as msg:
                     self.usage("invalid value for %s %r: %s" % (opt, arg, msg))
             if name and arg is not None:
                 if getattr(self, name) is not None:
@@ -280,12 +291,12 @@ class Options:
         # Process environment variables
         for envvar in self.environ_map.keys():
             name, handler = self.environ_map[envvar]
-            if os.environ.has_key(envvar):
+            if envvar in os.environ:
                 value = os.environ[envvar]
                 if handler is not None:
                     try:
                         value = handler(value)
-                    except ValueError, msg:
+                    except ValueError as msg:
                         self.usage("invalid environment value for %s %r: %s"
                                    % (envvar, value, msg))
                 if name and value is not None:
@@ -333,7 +344,7 @@ class Options:
             self.here = os.path.abspath(os.path.dirname(self.configfile))
         try:
             self.read_config(self.configfile)
-        except ValueError, msg:
+        except ValueError as msg:
             if do_usage:
                 # if this is not called from an RPC method, run usage and exit.
                 self.usage(str(msg))
@@ -450,10 +461,9 @@ class ServerOptions(Options):
         self.stdout.write('%s\n' % VERSION)
         self.exit(0)
 
-    def getLogger(self, filename, level, fmt, rotating=False, maxbytes=0,
-                  backups=0, stdout=False):
-        return loggers.getLogger(filename, level, fmt, rotating, maxbytes,
-                                 backups, stdout)
+    # TODO: not covered by any test, but used by dispatchers
+    def getLogger(self, *args, **kwargs):
+        return loggers.getLogger(*args, **kwargs)
 
     def default_configfile(self):
         if os.getuid() == 0:
@@ -475,7 +485,7 @@ class ServerOptions(Options):
         if self.user is not None:
             try:
                 uid = name_to_uid(self.user)
-            except ValueError, msg:
+            except ValueError as msg:
                 self.usage(msg) # invalid user
             self.uid = uid
             self.gid = gid_for_uid(uid)
@@ -534,32 +544,42 @@ class ServerOptions(Options):
         self.process_group_configs = new
 
     def read_config(self, fp):
-        # Clear parse warnings, since we may be re-reading the
+        # Clear parse messages, since we may be re-reading the
         # config a second time after a reload.
         self.parse_criticals = []
         self.parse_warnings = []
         self.parse_infos = []
 
         section = self.configroot.supervisord
+        need_close = False
         if not hasattr(fp, 'read'):
             if not self.exists(fp):
                 raise ValueError("could not find config file %s" % fp)
             try:
                 fp = self.open(fp, 'r')
+                need_close = True
             except (IOError, OSError):
                 raise ValueError("could not read config file %s" % fp)
+
         parser = UnhosedConfigParser()
         parser.expansions = self.environ_expansions
         try:
-            parser.readfp(fp)
-        except ConfigParser.ParsingError, why:
+            try:
+                parser.read_file(fp)
+            except AttributeError:
+                parser.readfp(fp)
+        except ConfigParser.ParsingError as why:
             raise ValueError(str(why))
+        finally:
+            if need_close:
+                fp.close()
 
         host_node_name = platform.node()
         expansions = {'here':self.here,
                       'host_node_name':host_node_name}
         expansions.update(self.environ_expansions)
         if parser.has_section('include'):
+            parser.expand_here(self.here)
             if not parser.has_option('include', 'files'):
                 raise ValueError(".ini file has [include] section, but no "
                 "files setting")
@@ -582,8 +602,12 @@ class ServerOptions(Options):
                         'Included extra file "%s" during parsing' % filename)
                     try:
                         parser.read(filename)
-                    except ConfigParser.ParsingError, why:
+                    except ConfigParser.ParsingError as why:
                         raise ValueError(str(why))
+                    else:
+                        parser.expand_here(
+                            os.path.abspath(os.path.dirname(filename))
+                        )
 
         sections = parser.sections()
         if not 'supervisord' in sections:
@@ -662,13 +686,19 @@ class ServerOptions(Options):
             group_processes = []
             for program in programs:
                 program_section = "program:%s" % program
-                if not program_section in all_sections:
+                fcgi_section = "fcgi-program:%s" % program
+                if not program_section in all_sections and not fcgi_section in all_sections:
                     raise ValueError(
-                        '[%s] names unknown program %s' % (section, program))
-                homogeneous_exclude.append(program_section)
-                processes = self.processes_from_section(parser, program_section,
-                                                        group_name,
-                                                        ProcessConfig)
+                        '[%s] names unknown program or fcgi-program %s' % (section, program))
+                if program_section in all_sections and fcgi_section in all_sections:
+                     raise ValueError(
+                        '[%s] name %s is ambiguous (exists as program and fcgi-program)' %
+                        (section, program))
+                section = program_section if program_section in all_sections else fcgi_section
+                homogeneous_exclude.append(section)
+                processes = self.processes_from_section(parser, section,
+                                                        group_name, ProcessConfig)
+
                 group_processes.extend(processes)
             groups.append(
                 ProcessGroupConfig(self, group_name, priority, group_processes)
@@ -757,6 +787,14 @@ class ServerOptions(Options):
             else:
                 proc_uid = name_to_uid(proc_user)
 
+            socket_backlog = get(section, 'socket_backlog', None)
+
+            if socket_backlog is not None:
+                socket_backlog = integer(socket_backlog)
+                if (socket_backlog < 1 or socket_backlog > 65535):
+                    raise ValueError('Invalid socket_backlog value %s'
+                                                            % socket_backlog)
+
             socket_owner = get(section, 'socket_owner', None)
             if socket_owner is not None:
                 try:
@@ -780,8 +818,9 @@ class ServerOptions(Options):
 
             try:
                 socket_config = self.parse_fcgi_socket(socket, proc_uid,
-                                                    socket_owner, socket_mode)
-            except ValueError, e:
+                                                    socket_owner, socket_mode,
+                                                    socket_backlog)
+            except ValueError as e:
                 raise ValueError('%s in [%s] socket' % (str(e), section))
 
             processes=self.processes_from_section(parser, section, program_name,
@@ -794,7 +833,8 @@ class ServerOptions(Options):
         groups.sort()
         return groups
 
-    def parse_fcgi_socket(self, sock, proc_uid, socket_owner, socket_mode):
+    def parse_fcgi_socket(self, sock, proc_uid, socket_owner, socket_mode,
+            socket_backlog):
         if sock.startswith('unix://'):
             path = sock[7:]
             #Check it's an absolute path
@@ -809,10 +849,11 @@ class ServerOptions(Options):
                     socket_owner = (proc_uid, gid_for_uid(proc_uid))
 
             if socket_mode is None:
-                socket_mode = int('700', 8)
+                socket_mode = 0o700
 
             return UnixStreamSocketConfig(path, owner=socket_owner,
-                                                mode=socket_mode)
+                                                mode=socket_mode,
+                                                backlog=socket_backlog)
 
         if socket_owner is not None or socket_mode is not None:
             raise ValueError("socket_owner and socket_mode params should"
@@ -822,7 +863,8 @@ class ServerOptions(Options):
         if m:
             host = m.group(1)
             port = int(m.group(2))
-            return InetStreamSocketConfig(host, port)
+            return InetStreamSocketConfig(host, port,
+                    backlog=socket_backlog)
 
         raise ValueError("Bad socket format %s", sock)
 
@@ -831,7 +873,7 @@ class ServerOptions(Options):
         try:
             return self._processes_from_section(
                 parser, section, group_name, klass)
-        except ValueError, e:
+        except ValueError as e:
             filename = parser.section_to_file.get(section, self.configfile)
             raise ValueError('%s in section %r (file: %r)'
                              % (e, section, filename))
@@ -863,7 +905,7 @@ class ServerOptions(Options):
         stopwaitsecs = integer(get(section, 'stopwaitsecs', 10))
         stopasgroup = boolean(get(section, 'stopasgroup', 'false'))
         killasgroup = boolean(get(section, 'killasgroup', stopasgroup))
-        exitcodes = list_of_exitcodes(get(section, 'exitcodes', '0,2'))
+        exitcodes = list_of_exitcodes(get(section, 'exitcodes', '0'))
         # see also redirect_stderr check in process_groups_from_parser()
         redirect_stderr = boolean(get(section, 'redirect_stderr','false'))
         numprocs = integer(get(section, 'numprocs', 1))
@@ -932,6 +974,10 @@ class ServerOptions(Options):
                 maxbytes = byte_size(get(section, mb_key, '50MB'))
                 logfiles[mb_key] = maxbytes
 
+                sy_key = '%s_syslog' % k
+                syslog = boolean(get(section, sy_key, False))
+                logfiles[sy_key] = syslog
+
                 if lf_val is Automatic and not maxbytes:
                     self.parse_warnings.append(
                         'For [%s], AUTO logging used for %s without '
@@ -969,11 +1015,13 @@ class ServerOptions(Options):
                 stdout_events_enabled = stdout_events,
                 stdout_logfile_backups=logfiles['stdout_logfile_backups'],
                 stdout_logfile_maxbytes=logfiles['stdout_logfile_maxbytes'],
+                stdout_syslog=logfiles['stdout_syslog'],
                 stderr_logfile=logfiles['stderr_logfile'],
                 stderr_capture_maxbytes = stderr_cmaxbytes,
                 stderr_events_enabled = stderr_events,
                 stderr_logfile_backups=logfiles['stderr_logfile_backups'],
                 stderr_logfile_maxbytes=logfiles['stderr_logfile_maxbytes'],
+                stderr_syslog=logfiles['stderr_syslog'],
                 stopsignal=stopsignal,
                 stopwaitsecs=stopwaitsecs,
                 stopasgroup=stopasgroup,
@@ -1058,7 +1106,7 @@ class ServerOptions(Options):
                 except (TypeError, ValueError):
                     raise ValueError('Invalid chmod value %s' % chmod)
             else:
-                chmod = int('700', 8)
+                chmod = 0o700
             config['chmod'] = chmod
             config['section'] = section
             configs.append(config)
@@ -1102,7 +1150,7 @@ class ServerOptions(Options):
         if self.directory:
             try:
                 os.chdir(self.directory)
-            except OSError, err:
+            except OSError as err:
                 self.logger.critical("can't chdir into %r: %s"
                                      % (self.directory, err))
             else:
@@ -1125,9 +1173,8 @@ class ServerOptions(Options):
     def write_pidfile(self):
         pid = os.getpid()
         try:
-            f = open(self.pidfile, 'w')
-            f.write('%s\n' % pid)
-            f.close()
+            with open(self.pidfile, 'w') as f:
+                f.write('%s\n' % pid)
         except (IOError, OSError):
             self.logger.critical('could not write pidfile %s' % self.pidfile)
         else:
@@ -1192,7 +1239,7 @@ class ServerOptions(Options):
         try:
             self.httpservers = self.make_http_servers(supervisord)
             self.unlink_socketfiles = True
-        except socket.error, why:
+        except socket.error as why:
             if why.args[0] == errno.EADDRINUSE:
                 self.usage('Another program is already listening on '
                            'a port that one of our HTTP servers is '
@@ -1206,7 +1253,7 @@ class ServerOptions(Options):
                 else:
                     self.usage('%s errno.%s (%d)' %
                                (help, errorname, why.args[0]))
-        except ValueError, why:
+        except ValueError as why:
             self.usage(why.args[0])
 
     def get_autochildlog_name(self, name, identifier, channel):
@@ -1262,7 +1309,7 @@ class ServerOptions(Options):
         # left zombies laying around.
         try:
             pid, sts = os.waitpid(-1, os.WNOHANG)
-        except OSError, exc:
+        except OSError as exc:
             code = exc.args[0]
             if code not in (errno.ECHILD, errno.EINTR):
                 self.logger.critical(
@@ -1412,16 +1459,18 @@ class ServerOptions(Options):
 
     def make_logger(self):
         # must be called after realize() and after supervisor does setuid()
-        format =  '%(asctime)s %(levelname)s %(message)s\n'
-        self.logger = loggers.getLogger(
+        format = '%(asctime)s %(levelname)s %(message)s\n'
+        self.logger = loggers.getLogger(self.loglevel)
+        if self.nodaemon:
+            loggers.handle_stdout(self.logger, format)
+        loggers.handle_file(
+            self.logger,
             self.logfile,
-            self.loglevel,
             format,
-            rotating=True,
+            rotating=not not self.logfile_maxbytes,
             maxbytes=self.logfile_maxbytes,
             backups=self.logfile_backups,
-            stdout = self.nodaemon,
-            )
+        )
         for msg in self.parse_criticals:
             self.logger.critical(msg)
         for msg in self.parse_warnings:
@@ -1452,7 +1501,7 @@ class ServerOptions(Options):
         return os.stat(filename)
 
     def write(self, fd, data):
-        return os.write(fd, data)
+        return os.write(fd, as_bytes(data))
 
     def execve(self, filename, argv, env):
         return os.execve(filename, argv, env)
@@ -1478,7 +1527,7 @@ class ServerOptions(Options):
     def get_path(self):
         """Return a list corresponding to $PATH, or a default."""
         path = ["/bin", "/usr/bin", "/usr/local/bin"]
-        if os.environ.has_key("PATH"):
+        if "PATH" in os.environ:
             p = os.environ["PATH"]
             if p:
                 path = p.split(os.pathsep)
@@ -1494,7 +1543,7 @@ class ServerOptions(Options):
         elif stat.S_ISDIR(st[stat.ST_MODE]):
             raise NotExecutable("command at %r is a directory" % filename)
 
-        elif not (stat.S_IMODE(st[stat.ST_MODE]) & int('111', 8)):
+        elif not (stat.S_IMODE(st[stat.ST_MODE]) & 0o111):
             raise NotExecutable("command at %r is not executable" % filename)
 
         elif not os.access(filename, os.X_OK):
@@ -1509,10 +1558,10 @@ class ServerOptions(Options):
     def readfd(self, fd):
         try:
             data = os.read(fd, 2 << 16) # 128K
-        except OSError, why:
+        except OSError as why:
             if why.args[0] not in (errno.EWOULDBLOCK, errno.EBADF, errno.EINTR):
                 raise
-            data = ''
+            data = b''
         return data
 
     def process_environment(self):
@@ -1523,7 +1572,7 @@ class ServerOptions(Options):
 
     def make_pipes(self, stderr=True):
         """ Create pipes for parent to child stdin/stdout/stderr
-        communications.  Open fd in nonblocking mode so we can read them
+        communications.  Open fd in non-blocking mode so we can read them
         in the mainloop without blocking.  If stderr is False, don't
         create a pipe for stderr. """
 
@@ -1607,18 +1656,26 @@ class ClientOptions(Options):
 
     def read_config(self, fp):
         section = self.configroot.supervisorctl
+        need_close = False
         if not hasattr(fp, 'read'):
             self.here = os.path.dirname(normalize_path(fp))
             if not self.exists(fp):
                 raise ValueError("could not find config file %s" % fp)
             try:
                 fp = self.open(fp, 'r')
+                need_close = True
             except (IOError, OSError):
                 raise ValueError("could not read config file %s" % fp)
+
         parser = UnhosedConfigParser()
         parser.expansions = self.environ_expansions
         parser.mysection = 'supervisorctl'
-        parser.readfp(fp)
+        try:
+            parser.read_file(fp)
+        except AttributeError:
+            parser.readfp(fp)
+        if need_close:
+            fp.close()
         sections = parser.sections()
         if not 'supervisorctl' in sections:
             raise ValueError('.ini file does not include supervisorctl section')
@@ -1653,8 +1710,8 @@ class ClientOptions(Options):
 
         return section
 
+    # TODO: not covered by any test, but used by supervisorctl
     def getServerProxy(self):
-        # mostly put here for unit testing
         return xmlrpclib.ServerProxy(
             # dumbass ServerProxy won't allow us to pass in a non-HTTP url,
             # so we fake the url we pass into it and always use the transport's
@@ -1671,14 +1728,29 @@ class UnhosedConfigParser(ConfigParser.RawConfigParser):
     mysection = 'supervisord'
 
     def __init__(self, *args, **kwargs):
+        # inline_comment_prefixes and strict were added in Python 3 but their
+        # defaults make RawConfigParser behave differently than it did on
+        # Python 2.  We make it work like 2 by default for backwards compat.
+        if not PY2:
+            if 'inline_comment_prefixes' not in kwargs:
+                kwargs['inline_comment_prefixes'] = (';', '#')
+
+            if 'strict' not in kwargs:
+                kwargs['strict'] = False
+
         ConfigParser.RawConfigParser.__init__(self, *args, **kwargs)
+
         self.section_to_file = {}
         self.expansions = {}
 
-    def read_string(self, s):
-        from StringIO import StringIO
-        s = StringIO(s)
-        return self.readfp(s)
+    def read_string(self, string, source='<string>'):
+        '''Parse configuration data from a string.  This is intended
+        to be used in tests only.  We add this method for Py 2/3 compat.'''
+        try:
+            return ConfigParser.RawConfigParser.read_string(
+                self, string, source) # Python 3.2 or later
+        except AttributeError:
+            return self.readfp(StringIO(string))
 
     def read(self, filenames, **kwargs):
         '''Attempt to read and parse a list of filenames, returning a list
@@ -1724,6 +1796,15 @@ class UnhosedConfigParser(ConfigParser.RawConfigParser):
         return self.saneget(self.mysection, option, default=default,
                             expansions=expansions, **kwargs)
 
+    def expand_here(self, here):
+        HERE_FORMAT = '%(here)s'
+        for section in self.sections():
+            for key, value in self.items(section):
+                if HERE_FORMAT in value:
+                    assert here is not None, "here has not been set to a path"
+                    value = value.replace(HERE_FORMAT, here)
+                    self.set(section, key, value)
+
 
 class Config(object):
     def __ne__(self, other):
@@ -1762,11 +1843,11 @@ class ProcessConfig(Config):
         'name', 'uid', 'command', 'directory', 'umask', 'priority',
         'autostart', 'autorestart', 'startsecs', 'startretries',
         'stdout_logfile', 'stdout_capture_maxbytes',
-        'stdout_events_enabled',
+        'stdout_events_enabled', 'stdout_syslog',
         'stdout_logfile_backups', 'stdout_logfile_maxbytes',
         'stderr_logfile', 'stderr_capture_maxbytes',
         'stderr_logfile_backups', 'stderr_logfile_maxbytes',
-        'stderr_events_enabled',
+        'stderr_events_enabled', 'stderr_syslog',
         'stopsignal', 'stopwaitsecs', 'stopasgroup', 'killasgroup',
         'exitcodes', 'redirect_stderr' ]
     optional_param_names = [ 'environment', 'serverurl' ]
@@ -1968,27 +2049,27 @@ def readFile(filename, offset, length):
     abslength = abs(length)
 
     try:
-        f = open(filename, 'rb')
-        if absoffset != offset:
-            # negative offset returns offset bytes from tail of the file
-            if length:
-                raise ValueError('BAD_ARGUMENTS')
-            f.seek(0, 2)
-            sz = f.tell()
-            pos = int(sz - absoffset)
-            if pos < 0:
-                pos = 0
-            f.seek(pos)
-            data = f.read(absoffset)
-        else:
-            if abslength != length:
-                raise ValueError('BAD_ARGUMENTS')
-            if length == 0:
-                f.seek(offset)
-                data = f.read()
+        with open(filename, 'rb') as f:
+            if absoffset != offset:
+                # negative offset returns offset bytes from tail of the file
+                if length:
+                    raise ValueError('BAD_ARGUMENTS')
+                f.seek(0, 2)
+                sz = f.tell()
+                pos = int(sz - absoffset)
+                if pos < 0:
+                    pos = 0
+                f.seek(pos)
+                data = f.read(absoffset)
             else:
-                sz = f.seek(offset)
-                data = f.read(length)
+                if abslength != length:
+                    raise ValueError('BAD_ARGUMENTS')
+                if length == 0:
+                    f.seek(offset)
+                    data = f.read()
+                else:
+                    f.seek(offset)
+                    data = f.read(length)
     except (OSError, IOError):
         raise ValueError('FAILED')
 
@@ -2002,33 +2083,34 @@ def tailFile(filename, offset, length):
     bytes are not available, as many bytes as are available are returned.
     """
 
-    overflow = False
     try:
-        f = open(filename, 'rb')
-        f.seek(0, 2)
-        sz = f.tell()
+        with open(filename, 'rb') as f:
+            overflow = False
+            f.seek(0, 2)
+            sz = f.tell()
 
-        if sz > (offset + length):
-            overflow = True
-            offset   = sz - 1
+            if sz > (offset + length):
+                overflow = True
+                offset = sz - 1
 
-        if (offset + length) > sz:
-            if (offset > (sz - 1)):
+            if (offset + length) > sz:
+                if offset > (sz - 1):
+                    length = 0
+                offset = sz - length
+
+            if offset < 0:
+                offset = 0
+            if length < 0:
                 length = 0
-            offset = sz - length
 
-        if offset < 0: offset = 0
-        if length < 0: length = 0
-
-        if length == 0:
-            data = ''
-        else:
-            f.seek(offset)
-            data = f.read(length)
-
-        offset = sz
-        return [data, offset, overflow]
+            if length == 0:
+                data = b''
+            else:
+                f.seek(offset)
+                data = f.read(length)
 
+            offset = sz
+            return [as_string(data), offset, overflow]
     except (OSError, IOError):
         return ['', offset, False]
 
@@ -2104,14 +2186,14 @@ class SignalReceiver:
 def expand(s, expansions, name):
     try:
         return s % expansions
-    except KeyError, ex:
+    except KeyError as ex:
         available = list(expansions.keys())
         available.sort()
         raise ValueError(
             'Format string %r for %r contains names (%s) which cannot be '
             'expanded. Available names: %s' %
             (s, name, str(ex), ", ".join(available)))
-    except Exception, ex:
+    except Exception as ex:
         raise ValueError(
             'Format string %r for %r is badly formatted: %s' %
             (s, name, str(ex))
