@@ -1,10 +1,15 @@
 import os
-import sys
 import time
 import errno
 import shlex
 import traceback
 import signal
+from functools import total_ordering
+
+from supervisor.compat import maxint
+from supervisor.compat import as_bytes
+from supervisor.compat import as_string
+from supervisor.compat import PY2
 
 from supervisor.medusa import asyncore_25 as asyncore
 
@@ -25,7 +30,8 @@ from supervisor.datatypes import RestartUnconditionally
 
 from supervisor.socket_manager import SocketManager
 
-class Subprocess:
+@total_ordering
+class Subprocess(object):
 
     """A class to manage a subprocess."""
 
@@ -42,11 +48,11 @@ class Subprocess:
     delay = 0 # If nonzero, delay starting or killing until this time
     administrative_stop = False # true if process has been stopped by an admin
     system_stop = False # true if process has been stopped by the system
-    killing = False # true if are trying to kill this process
+    killing = False # true if we are trying to kill this process
     backoff = 0 # backoff counter (to startretries)
-    dispatchers = None # asnycore output dispatchers (keyed by fd)
+    dispatchers = None # asyncore output dispatchers (keyed by fd)
     pipes = None # map of channel name to file descriptor #
-    exitstatus = None # status attached to dead process by finsh()
+    exitstatus = None # status attached to dead process by finish()
     spawnerr = None # error message attached by spawn() if any
     group = None # ProcessGroup instance if process is in the group
 
@@ -102,7 +108,7 @@ class Subprocess:
         if not """
         try:
             commandargs = shlex.split(self.config.command)
-        except ValueError, e:
+        except ValueError as e:
             raise BadCommand("can't parse command %r: %s" % \
                 (self.config.command, str(e)))
 
@@ -119,7 +125,7 @@ class Subprocess:
                 st = None
 
         else:
-            path = self.config.options.get_path()
+            path = self.config.get_path()
             found = None
             st = None
             for dir in path:
@@ -166,7 +172,7 @@ class Subprocess:
 
         if new_state == ProcessStates.BACKOFF:
             now = time.time()
-            self.backoff = self.backoff + 1
+            self.backoff += 1
             self.delay = now + self.backoff
 
         self.state = new_state
@@ -175,8 +181,9 @@ class Subprocess:
         if self.state not in states:
             current_state = getProcessStateDescription(self.state)
             allowable_states = ' '.join(map(getProcessStateDescription, states))
+            processname = as_string(self.config.name)
             raise AssertionError('Assertion failed for %s: %s not in %s' %  (
-                self.config.name, current_state, allowable_states))
+                processname, current_state, allowable_states))
 
     def record_spawnerr(self, msg):
         self.spawnerr = msg
@@ -188,9 +195,10 @@ class Subprocess:
         Return the process id.  If the fork() call fails, return None.
         """
         options = self.config.options
+        processname = as_string(self.config.name)
 
         if self.pid:
-            msg = 'process %r already running' % self.config.name
+            msg = 'process \'%s\' already running' % processname
             options.logger.warn(msg)
             return
 
@@ -209,7 +217,7 @@ class Subprocess:
 
         try:
             filename, argv = self.get_execv_args()
-        except ProcessException, what:
+        except ProcessException as what:
             self.record_spawnerr(what.args[0])
             self._assertInState(ProcessStates.STARTING)
             self.change_state(ProcessStates.BACKOFF)
@@ -217,14 +225,14 @@ class Subprocess:
 
         try:
             self.dispatchers, self.pipes = self.config.make_dispatchers(self)
-        except (OSError, IOError), why:
+        except (OSError, IOError) as why:
             code = why.args[0]
             if code == errno.EMFILE:
                 # too many file descriptors open
-                msg = 'too many open files to spawn %r' % self.config.name
+                msg = 'too many open files to spawn \'%s\'' % processname
             else:
-                msg = 'unknown error making dispatchers for %r: %s' % (
-                      self.config.name, errno.errorcode.get(code, code))
+                msg = 'unknown error making dispatchers for \'%s\': %s' % (
+                      processname, errno.errorcode.get(code, code))
             self.record_spawnerr(msg)
             self._assertInState(ProcessStates.STARTING)
             self.change_state(ProcessStates.BACKOFF)
@@ -232,15 +240,15 @@ class Subprocess:
 
         try:
             pid = options.fork()
-        except OSError, why:
+        except OSError as why:
             code = why.args[0]
             if code == errno.EAGAIN:
                 # process table full
-                msg  = ('Too many processes in process table to spawn %r' %
-                        self.config.name)
+                msg  = ('Too many processes in process table to spawn \'%s\'' %
+                        processname)
             else:
-                msg = 'unknown error during fork for %r: %s' % (
-                      self.config.name, errno.errorcode.get(code, code))
+                msg = 'unknown error during fork for \'%s\': %s' % (
+                      processname, errno.errorcode.get(code, code))
             self.record_spawnerr(msg)
             self._assertInState(ProcessStates.STARTING)
             self.change_state(ProcessStates.BACKOFF)
@@ -259,7 +267,7 @@ class Subprocess:
         self.pid = pid
         options = self.config.options
         options.close_child_pipes(self.pipes)
-        options.logger.info('spawned: %r with pid %s' % (self.config.name, pid))
+        options.logger.info('spawned: \'%s\' with pid %s' % (as_string(self.config.name), pid))
         self.spawnerr = None
         self.delay = time.time() + self.config.startsecs
         options.pidhistory[pid] = self
@@ -315,11 +323,11 @@ class Subprocess:
                 env.update(self.config.environment)
 
             # change directory
+            cwd = self.config.directory
             try:
-                cwd = self.config.directory
                 if cwd is not None:
                     options.chdir(cwd)
-            except OSError, why:
+            except OSError as why:
                 code = errno.errorcode.get(why.args[0], why.args[0])
                 msg = "couldn't chdir to %s: %s\n" % (cwd, code)
                 options.write(2, "supervisor: " + msg)
@@ -330,7 +338,7 @@ class Subprocess:
                 if self.config.umask is not None:
                     options.setumask(self.config.umask)
                 options.execve(filename, argv, env)
-            except OSError, why:
+            except OSError as why:
                 code = errno.errorcode.get(why.args[0], why.args[0])
                 msg = "couldn't exec %s: %s\n" % (argv[0], code)
                 options.write(2, "supervisor: " + msg)
@@ -347,6 +355,28 @@ class Subprocess:
             options.write(2, "supervisor: child process was not spawned\n")
             options._exit(127) # exit process with code for spawn failure
 
+    def _check_and_adjust_for_system_clock_rollback(self, test_time):
+        """
+        Check if system clock has rolled backward beyond test_time. If so, set
+        affected timestamps to test_time.
+        """
+        if self.state == ProcessStates.STARTING:
+            if test_time < self.laststart:
+                self.laststart = test_time;
+            if self.delay > 0 and test_time < (self.delay - self.config.startsecs):
+                self.delay = test_time + self.config.startsecs
+        elif self.state == ProcessStates.RUNNING:
+            if test_time > self.laststart and test_time < (self.laststart + self.config.startsecs):
+                self.laststart = test_time - self.config.startsecs
+        elif self.state == ProcessStates.STOPPING:
+            if test_time < self.laststopreport:
+                self.laststopreport = test_time;
+            if self.delay > 0 and test_time < (self.delay - self.config.stopwaitsecs):
+                self.delay = test_time + self.config.stopwaitsecs
+        elif self.state == ProcessStates.BACKOFF:
+            if self.delay > 0 and test_time < (self.delay - self.backoff):
+                self.delay = test_time + self.backoff
+
     def stop(self):
         """ Administrative stop """
         self.administrative_stop = True
@@ -357,9 +387,12 @@ class Subprocess:
         """ Log a 'waiting for x to stop' message with throttling. """
         if self.state == ProcessStates.STOPPING:
             now = time.time()
+
+            self._check_and_adjust_for_system_clock_rollback(now)
+
             if now > (self.laststopreport + 2): # every 2 seconds
                 self.config.options.logger.info(
-                    'waiting for %s to stop' % self.config.name)
+                    'waiting for %s to stop' % as_string(self.config.name))
                 self.laststopreport = now
 
     def give_up(self):
@@ -378,22 +411,26 @@ class Subprocess:
         now = time.time()
         options = self.config.options
 
-        # Properly stop processes in BACKOFF state.
+        processname = as_string(self.config.name)
+        # If the process is in BACKOFF and we want to stop or kill it, then
+        # BACKOFF -> STOPPED.  This is needed because if startretries is a
+        # large number and the process isn't starting successfully, the stop
+        # request would be blocked for a long time waiting for the retries.
         if self.state == ProcessStates.BACKOFF:
             msg = ("Attempted to kill %s, which is in BACKOFF state." %
-                   (self.config.name))
+                   processname)
             options.logger.debug(msg)
             self.change_state(ProcessStates.STOPPED)
             return None
 
         if not self.pid:
             msg = ("attempted to kill %s with sig %s but it wasn't running" %
-                   (self.config.name, signame(sig)))
+                   (processname, signame(sig)))
             options.logger.debug(msg)
             return msg
 
-        #If we're in the stopping state, then we've already sent the stop
-        #signal and this is the kill signal
+        # If we're in the stopping state, then we've already sent the stop
+        # signal and this is the kill signal
         if self.state == ProcessStates.STOPPING:
             killasgroup = self.config.killasgroup
         else:
@@ -404,7 +441,7 @@ class Subprocess:
             as_group = "process group "
 
         options.logger.debug('killing %s (pid %s) %swith signal %s'
-                             % (self.config.name,
+                             % (processname,
                                 self.pid,
                                 as_group,
                                 signame(sig))
@@ -415,7 +452,8 @@ class Subprocess:
         self.delay = now + self.config.stopwaitsecs
         # we will already be in the STOPPING state if we're doing a
         # SIGKILL as a result of overrunning stopwaitsecs
-        self._assertInState(ProcessStates.RUNNING,ProcessStates.STARTING,
+        self._assertInState(ProcessStates.RUNNING,
+                            ProcessStates.STARTING,
                             ProcessStates.STOPPING)
         self.change_state(ProcessStates.STOPPING)
 
@@ -428,7 +466,7 @@ class Subprocess:
             options.kill(pid, sig)
         except:
             tb = traceback.format_exc()
-            msg = 'unknown problem killing %s (%s):%s' % (self.config.name,
+            msg = 'unknown problem killing %s (%s):%s' % (processname,
                                                           self.pid, tb)
             options.logger.critical(msg)
             self.change_state(ProcessStates.UNKNOWN)
@@ -446,19 +484,21 @@ class Subprocess:
         if an error occurred or if the subprocess is not running.
         """
         options = self.config.options
+        processname = as_string(self.config.name)
         if not self.pid:
             msg = ("attempted to send %s sig %s but it wasn't running" %
-                   (self.config.name, signame(sig)))
+                   (processname, signame(sig)))
             options.logger.debug(msg)
             return msg
 
         options.logger.debug('sending %s (pid %s) sig %s'
-                             % (self.config.name,
+                             % (processname,
                                 self.pid,
                                 signame(sig))
                              )
 
-        self._assertInState(ProcessStates.RUNNING,ProcessStates.STARTING,
+        self._assertInState(ProcessStates.RUNNING,
+                            ProcessStates.STARTING,
                             ProcessStates.STOPPING)
 
         try:
@@ -466,7 +506,7 @@ class Subprocess:
         except:
             tb = traceback.format_exc()
             msg = 'unknown problem sending sig %s (%s):%s' % (
-                                self.config.name, self.pid, tb)
+                                processname, self.pid, tb)
             options.logger.critical(msg)
             self.change_state(ProcessStates.UNKNOWN)
             self.pid = 0
@@ -482,17 +522,20 @@ class Subprocess:
         es, msg = decode_wait_status(sts)
 
         now = time.time()
+
+        self._check_and_adjust_for_system_clock_rollback(now)
+
         self.laststop = now
-        processname = self.config.name
+        processname = as_string(self.config.name)
 
         if now > self.laststart:
             too_quickly = now - self.laststart < self.config.startsecs
         else:
             too_quickly = False
             self.config.options.logger.warn(
-                "process %r (%s) laststart time is in the future, don't "
+                "process \'%s\' (%s) laststart time is in the future, don't "
                 "know how long process was running so assuming it did "
-                "not exit too quickly" % (self.config.name, self.pid))
+                "not exit too quickly" % (processname, self.pid))
 
         exit_expected = es in self.config.exitcodes
 
@@ -561,17 +604,25 @@ class Subprocess:
     def set_uid(self):
         if self.config.uid is None:
             return
-        msg = self.config.options.dropPrivileges(self.config.uid)
+        msg = self.config.options.drop_privileges(self.config.uid)
         return msg
 
-    def __cmp__(self, other):
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
+    def __eq__(self, other):
         # sort by priority
-        return cmp(self.config.priority, other.config.priority)
+        return self.config.priority == other.config.priority
 
     def __repr__(self):
+        # repr can't return anything other than a native string,
+        # but the name might be unicode - a problem on Python 2.
+        name = self.config.name
+        if PY2:
+            name = as_string(name).encode('unicode-escape')
         return '<Subprocess at %s with name %s in state %s>' % (
             id(self),
-            self.config.name,
+            name,
             getProcessStateDescription(self.get_state()))
 
     def get_state(self):
@@ -581,6 +632,8 @@ class Subprocess:
         now = time.time()
         state = self.state
 
+        self._check_and_adjust_for_system_clock_rollback(now)
+
         logger = self.config.options.logger
 
         if self.config.options.mood > SupervisorStates.RESTARTING:
@@ -604,6 +657,7 @@ class Subprocess:
                         # BACKOFF -> STARTING
                         self.spawn()
 
+        processname = as_string(self.config.name)
         if state == ProcessStates.STARTING:
             if now - self.laststart > self.config.startsecs:
                 # STARTING -> RUNNING if the proc has started
@@ -616,7 +670,7 @@ class Subprocess:
                 msg = (
                     'entered RUNNING state, process has stayed up for '
                     '> than %s seconds (startsecs)' % self.config.startsecs)
-                logger.info('success: %s %s' % (self.config.name, msg))
+                logger.info('success: %s %s' % (processname, msg))
 
         if state == ProcessStates.BACKOFF:
             if self.backoff > self.config.startretries:
@@ -625,7 +679,7 @@ class Subprocess:
                 self.give_up()
                 msg = ('entered FATAL state, too many start retries too '
                        'quickly')
-                logger.info('gave up: %s %s' % (self.config.name, msg))
+                logger.info('gave up: %s %s' % (processname, msg))
 
         elif state == ProcessStates.STOPPING:
             time_left = self.delay - now
@@ -634,8 +688,8 @@ class Subprocess:
                 # sigkill.  if this doesn't kill it, the process will be stuck
                 # in the STOPPING state forever.
                 self.config.options.logger.warn(
-                    'killing %r (%s) with SIGKILL' % (self.config.name,
-                                                      self.pid))
+                    'killing \'%s\' (%s) with SIGKILL' % (processname,
+                                                          self.pid))
                 self.kill(signal.SIGKILL)
 
 class FastCGISubprocess(Subprocess):
@@ -699,7 +753,8 @@ class FastCGISubprocess(Subprocess):
         for i in range(3, options.minfds):
             options.close_fd(i)
 
-class ProcessGroupBase:
+@total_ordering
+class ProcessGroupBase(object):
     def __init__(self, config):
         self.config = config
         self.processes = {}
@@ -707,12 +762,20 @@ class ProcessGroupBase:
             self.processes[pconfig.name] = pconfig.make_process(self)
 
 
-    def __cmp__(self, other):
-        return cmp(self.config.priority, other.config.priority)
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
+    def __eq__(self, other):
+        return self.config.priority == other.config.priority
 
     def __repr__(self):
+        # repr can't return anything other than a native string,
+        # but the name might be unicode - a problem on Python 2.
+        name = self.config.name
+        if PY2:
+            name = as_string(name).encode('unicode-escape')
         return '<%s instance at %s named %s>' % (self.__class__, id(self),
-                                                 self.config.name)
+                                                 name)
 
     def removelogs(self):
         for process in self.processes.values():
@@ -723,7 +786,7 @@ class ProcessGroupBase:
             process.reopenlogs()
 
     def stop_all(self):
-        processes = self.processes.values()
+        processes = list(self.processes.values())
         processes.sort()
         processes.reverse() # stop in desc priority order
 
@@ -750,6 +813,9 @@ class ProcessGroupBase:
             dispatchers.update(process.dispatchers)
         return dispatchers
 
+    def before_remove(self):
+        pass
+
 class ProcessGroup(ProcessGroupBase):
     def transition(self):
         for proc in self.processes.values():
@@ -766,19 +832,20 @@ class FastCGIProcessGroup(ProcessGroup):
         # to fail early during start up if there is a config error
         try:
             self.socket_manager.get_socket()
-        except Exception, e:
-            raise ValueError('Could not create FastCGI socket %s: %s' % (self.socket_manager.config(), e))
+        except Exception as e:
+            raise ValueError(
+                'Could not create FastCGI socket %s: %s' % (
+                    self.socket_manager.config(), e)
+                )
 
 class EventListenerPool(ProcessGroupBase):
     def __init__(self, config):
         ProcessGroupBase.__init__(self, config)
         self.event_buffer = []
-        for event_type in self.config.pool_events:
-            events.subscribe(event_type, self._acceptEvent)
-        events.subscribe(events.EventRejectedEvent, self.handle_rejected)
         self.serial = -1
         self.last_dispatch = 0
         self.dispatch_throttle = 0 # in seconds: .00195 is an interesting one
+        self._subscribe()
 
     def handle_rejected(self, event):
         process = event.process
@@ -800,10 +867,19 @@ class EventListenerPool(ProcessGroupBase):
         if dispatch_capable:
             if self.dispatch_throttle:
                 now = time.time()
+
+                if now < self.last_dispatch:
+                    # The system clock appears to have moved backward
+                    # Reset self.last_dispatch accordingly
+                    self.last_dispatch = now;
+
                 if now - self.last_dispatch < self.dispatch_throttle:
                     return
             self.dispatch()
 
+    def before_remove(self):
+        self._unsubscribe()
+
     def dispatch(self):
         while self.event_buffer:
             # dispatch the oldest event
@@ -818,17 +894,20 @@ class EventListenerPool(ProcessGroupBase):
 
     def _acceptEvent(self, event, head=False):
         # events are required to be instances
-        # this has a side effect to fail with an attribute error on 'old style' classes
+        # this has a side effect to fail with an attribute error on 'old style'
+        # classes
+        processname = as_string(self.config.name)
         if not hasattr(event, 'serial'):
             event.serial = new_serial(GlobalSerial)
         if not hasattr(event, 'pool_serials'):
             event.pool_serials = {}
-        if not event.pool_serials.has_key(self.config.name):
+        if self.config.name not in event.pool_serials:
             event.pool_serials[self.config.name] = new_serial(self)
         else:
             self.config.options.logger.debug(
-                'rebuffering event %s for pool %s (bufsize %s)' % (
-                (event.serial, self.config.name, len(self.event_buffer))))
+                'rebuffering event %s for pool %s (buf size=%d, max=%d)' % (
+                (event.serial, processname, len(self.event_buffer),
+                self.config.buffer_size)))
 
         if len(self.event_buffer) >= self.config.buffer_size:
             if self.event_buffer:
@@ -836,7 +915,7 @@ class EventListenerPool(ProcessGroupBase):
                 discarded_event = self.event_buffer.pop(0)
                 self.config.options.logger.error(
                     'pool %s event buffer overflowed, discarding event %s' % (
-                    (self.config.name, discarded_event.serial)))
+                    (processname, discarded_event.serial)))
         if head:
             self.event_buffer.insert(0, event)
         else:
@@ -849,23 +928,29 @@ class EventListenerPool(ProcessGroupBase):
             if process.state != ProcessStates.RUNNING:
                 continue
             if process.listener_state == EventListenerStates.READY:
-                payload = str(event)
+                processname = as_string(process.config.name)
+                payload = event.payload()
                 try:
                     event_type = event.__class__
                     serial = event.serial
                     envelope = self._eventEnvelope(event_type, serial,
                                                    pool_serial, payload)
-                    process.write(envelope)
-                except OSError, why:
+                    process.write(as_bytes(envelope))
+                except OSError as why:
                     if why.args[0] != errno.EPIPE:
                         raise
+
+                    self.config.options.logger.debug(
+                        'epipe occurred while sending event %s '
+                        'to listener %s, listener state unchanged' % (
+                        event.serial, processname))
                     continue
 
                 process.listener_state = EventListenerStates.BUSY
                 process.event = event
                 self.config.options.logger.debug(
                     'event %s sent to listener %s' % (
-                    event.serial, process.config.name))
+                    event.serial, processname))
                 return True
 
         return False
@@ -887,17 +972,25 @@ class EventListenerPool(ProcessGroupBase):
                 'pool:%(pool_name)s poolserial:%(pool_serial)s '
                 'eventname:%(event_name)s len:%(len)s\n%(payload)s' % D)
 
-class GlobalSerial:
+    def _subscribe(self):
+        for event_type in self.config.pool_events:
+            events.subscribe(event_type, self._acceptEvent)
+        events.subscribe(events.EventRejectedEvent, self.handle_rejected)
+
+    def _unsubscribe(self):
+        for event_type in self.config.pool_events:
+            events.unsubscribe(event_type, self._acceptEvent)
+        events.unsubscribe(events.EventRejectedEvent, self.handle_rejected)
+
+
+class GlobalSerial(object):
     def __init__(self):
         self.serial = -1
 
 GlobalSerial = GlobalSerial() # singleton
 
 def new_serial(inst):
-    if inst.serial == sys.maxint:
+    if inst.serial == maxint:
         inst.serial = -1
     inst.serial += 1
     return inst.serial
-
-
-
