@@ -1,6 +1,25 @@
 _NOW = 1151365354
 _TIMEFORMAT = '%b %d %I:%M %p'
 
+from functools import total_ordering
+
+from supervisor.compat import Fault
+from supervisor.compat import as_bytes
+
+# mock is imported here for py2/3 compat.  we only declare mock as a dependency
+# via tests_require so it is not available on all supervisor installs.  the
+# modules imported in supervisor.compat must always be available.
+
+try: # pragma: no cover
+    from unittest.mock import Mock, patch, sentinel
+except ImportError: # pragma: no cover
+    from mock import Mock, patch, sentinel
+
+try: # pragma: no cover
+    import unittest.mock as mock
+except ImportError: # pragma: no cover
+    import mock
+
 class DummyOptions:
 
     make_pipes_error = None
@@ -152,9 +171,7 @@ class DummyOptions:
     def make_pipes(self, stderr=True):
         if self.make_pipes_error:
             raise OSError(self.make_pipes_error)
-        pipes = {}
-        pipes['child_stdin'], pipes['stdin'] = (3, 4)
-        pipes['stdout'], pipes['child_stdout'] = (5, 6)
+        pipes = {'child_stdin': 3, 'stdin': 4, 'stdout': 5, 'child_stdout': 6}
         if stderr:
             pipes['stderr'], pipes['child_stderr'] = (7, 8)
         else:
@@ -246,6 +263,8 @@ class DummyOptions:
         self.umaskset = mask
 
 class DummyLogger:
+    level = None
+
     def __init__(self):
         self.reopened = False
         self.removed = False
@@ -263,6 +282,9 @@ class DummyLogger:
             msg = msg % kw
         self.data.append(msg)
 
+    def addHandler(self, handler):
+        handler.close()
+
     def reopen(self):
         self.reopened = True
     def close(self):
@@ -334,6 +356,9 @@ class DummySocketConfig:
     def __ne__(self, other):
         return not self.__eq__(other)
 
+    def get_backlog(self):
+        return 128
+
     def create_and_bind(self):
         return DummySocket(self.fd)
 
@@ -347,7 +372,8 @@ class DummySocketManager:
     def get_socket(self):
         return DummySocket(self._config.fd)
 
-class DummyProcess:
+@total_ordering
+class DummyProcess(object):
     # Initial state; overridden by instance variables
     pid = 0 # Subprocess pid; 0 when not running
     laststart = 0 # Last time the subprocess was started; 0 if never
@@ -386,11 +412,11 @@ class DummyProcess:
         self.error_at_clear = False
         self.killed_with = None
         self.drained = False
-        self.stdout_buffer = ''
-        self.stderr_buffer = ''
-        self.stdout_logged = ''
-        self.stderr_logged = ''
-        self.stdin_buffer = ''
+        self.stdout_buffer = as_bytes('')
+        self.stderr_buffer = as_bytes('')
+        self.stdout_logged = as_bytes('')
+        self.stderr_logged = as_bytes('')
+        self.stdin_buffer = as_bytes('')
         self.pipes = {}
         self.rpipes = {}
         self.dispatchers = {}
@@ -437,9 +463,6 @@ class DummyProcess:
     def drain(self):
         self.drained = True
 
-    def __cmp__(self, other):
-        return cmp(self.config.priority, other.config.priority)
-
     def readable_fds(self):
         return []
 
@@ -479,6 +502,12 @@ class DummyProcess:
     def transition(self):
         self.transitioned = True
 
+    def __eq__(self, other):
+        return self.config.priority == other.config.priority
+
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
 class DummyPConfig:
     def __init__(self, options, name, command, directory=None, umask=None,
                  priority=999, autostart=True,
@@ -486,12 +515,14 @@ class DummyPConfig:
                  uid=None, stdout_logfile=None, stdout_capture_maxbytes=0,
                  stdout_events_enabled=False,
                  stdout_logfile_backups=0, stdout_logfile_maxbytes=0,
+                 stdout_syslog=False,
                  stderr_logfile=None, stderr_capture_maxbytes=0,
                  stderr_events_enabled=False,
                  stderr_logfile_backups=0, stderr_logfile_maxbytes=0,
+                 stderr_syslog=False,
                  redirect_stderr=False,
                  stopsignal=None, stopwaitsecs=10, stopasgroup=False, killasgroup=False,
-                 exitcodes=(0,2), environment=None, serverurl=None):
+                 exitcodes=(0,), environment=None, serverurl=None):
         self.options = options
         self.name = name
         self.command = command
@@ -506,11 +537,13 @@ class DummyPConfig:
         self.stdout_events_enabled = stdout_events_enabled
         self.stdout_logfile_backups = stdout_logfile_backups
         self.stdout_logfile_maxbytes = stdout_logfile_maxbytes
+        self.stdout_syslog = stdout_syslog
         self.stderr_logfile = stderr_logfile
         self.stderr_capture_maxbytes = stderr_capture_maxbytes
         self.stderr_events_enabled = stderr_events_enabled
         self.stderr_logfile_backups = stderr_logfile_backups
         self.stderr_logfile_maxbytes = stderr_logfile_maxbytes
+        self.stderr_syslog = stderr_syslog
         self.redirect_stderr = redirect_stderr
         if stopsignal is None:
             import signal
@@ -566,10 +599,9 @@ def makeExecutable(file, substitutions=None):
         data = data.replace('<<%s>>' % key.upper(), substitutions[key])
 
     tmpnam = tempfile.mktemp(prefix=last)
-    f = open(tmpnam, 'w')
-    f.write(data)
-    f.close()
-    os.chmod(tmpnam, int('755', 8))
+    with open(tmpnam, 'w') as f:
+        f.write(data)
+    os.chmod(tmpnam, 0o755)
     return tmpnam
 
 def makeSpew(unkillable=False):
@@ -639,7 +671,10 @@ class DummyRequest(object):
         del self.headers[header]
 
     def has_key(self, header):
-        return self.headers.has_key(header)
+        return header in self.headers
+
+    def __contains__(self, item):
+        return item in self.headers
 
     def done(self):
         self._done = True
@@ -725,41 +760,42 @@ class DummySupervisorRPCNamespace:
     def getPID(self):
         return 42
 
-    def readProcessStdoutLog(self, name, offset, length):
+    def _read_log(self, channel, name, offset, length):
         from supervisor import xmlrpc
-        import xmlrpclib
         if name == 'BAD_NAME':
-            raise xmlrpclib.Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
+            raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         elif name == 'FAILED':
-            raise xmlrpclib.Fault(xmlrpc.Faults.FAILED, 'FAILED')
+            raise Fault(xmlrpc.Faults.FAILED, 'FAILED')
         elif name == 'NO_FILE':
-            raise xmlrpclib.Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
-        a = 'output line\n' * 10
+            raise Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
+        a = (channel + ' line\n') * 10
         return a[offset:]
 
+    def readProcessStdoutLog(self, name, offset, length):
+        return self._read_log('stdout', name, offset, length)
     readProcessLog = readProcessStdoutLog
-    readProcessStderrLog = readProcessStdoutLog
+
+    def readProcessStderrLog(self, name, offset, length):
+        return self._read_log('stderr', name, offset, length)
 
     def getAllProcessInfo(self):
         return self.all_process_info
 
     def getProcessInfo(self, name):
         from supervisor import xmlrpc
-        import xmlrpclib
         for i in self.all_process_info:
             if i['name']==name:
                 info=i
                 return info
         if name == 'BAD_NAME':
-            raise xmlrpclib.Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
+            raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         if name == 'FAILED':
-            raise xmlrpclib.Fault(xmlrpc.Faults.FAILED, 'FAILED')
+            raise Fault(xmlrpc.Faults.FAILED, 'FAILED')
         if name == 'NO_FILE':
-            raise xmlrpclib.Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
+            raise Fault(xmlrpc.Faults.NO_FILE, 'NO_FILE')
 
     def startProcess(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME:BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME:BAD_NAME')
         if name == 'BAD_NAME':
@@ -779,7 +815,7 @@ class DummySupervisorRPCNamespace:
 
     def startProcessGroup(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
+        from supervisor.compat import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -807,7 +843,7 @@ class DummySupervisorRPCNamespace:
 
     def stopProcessGroup(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
+        from supervisor.compat import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -821,7 +857,6 @@ class DummySupervisorRPCNamespace:
 
     def stopProcess(self, name):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME:BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME:BAD_NAME')
         if name == 'BAD_NAME':
@@ -850,7 +885,6 @@ class DummySupervisorRPCNamespace:
         if self._restartable:
             self._restarted = True
             return
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         raise Fault(xmlrpc.Faults.SHUTDOWN_STATE, '')
 
@@ -858,7 +892,6 @@ class DummySupervisorRPCNamespace:
         if self._restartable:
             self._shutdown = True
             return
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         raise Fault(xmlrpc.Faults.SHUTDOWN_STATE, '')
 
@@ -866,7 +899,6 @@ class DummySupervisorRPCNamespace:
         return [[['added'], ['changed'], ['removed']]]
 
     def addProcessGroup(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'ALREADY_ADDED':
             raise Fault(xmlrpc.Faults.ALREADY_ADDED, '')
@@ -882,7 +914,6 @@ class DummySupervisorRPCNamespace:
             self.processes = [name]
 
     def removeProcessGroup(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'STILL_RUNNING':
             raise Fault(xmlrpc.Faults.STILL_RUNNING, '')
@@ -893,7 +924,6 @@ class DummySupervisorRPCNamespace:
         self.processes.remove(name)
 
     def clearProcessStdoutLog(self, name):
-        from xmlrpclib import Fault
         from supervisor import xmlrpc
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
@@ -931,13 +961,11 @@ class DummySupervisorRPCNamespace:
 
     def readLog(self, whence, offset):
         if self._readlog_error:
-            from xmlrpclib import Fault
             raise Fault(self._readlog_error, '')
         return 'mainlogdata'
 
     def signalProcessGroup(self, name, signal):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if name == 'BAD_NAME':
             raise Fault(xmlrpc.Faults.BAD_NAME, 'BAD_NAME')
         return [
@@ -953,7 +981,6 @@ class DummySupervisorRPCNamespace:
 
     def signalProcess(self, name, signal):
         from supervisor import xmlrpc
-        from xmlrpclib import Fault
         if signal == 'BAD_SIGNAL':
             raise Fault(xmlrpc.Faults.BAD_SIGNAL, 'BAD_SIGNAL')
         if name == 'BAD_NAME:BAD_NAME':
@@ -1011,7 +1038,8 @@ class DummyFCGIGroupConfig(DummyPGroupConfig):
         DummyPGroupConfig.__init__(self, options, name, priority, pconfigs)
         self.socket_config = socket_config
 
-class DummyProcessGroup:
+@total_ordering
+class DummyProcessGroup(object):
     def __init__(self, config):
         self.config = config
         self.transitioned = False
@@ -1035,6 +1063,15 @@ class DummyProcessGroup:
     def get_dispatchers(self):
         return self.dispatchers
 
+    def __lt__(self, other):
+        return self.config.priority < other.config.priority
+
+    def __eq__(self, other):
+        return self.config.priority == other.config.priority
+
+    def reopenlogs(self):
+        self.logs_reopened = True
+
 class DummyFCGIProcessGroup(DummyProcessGroup):
 
     def __init__(self, config):
@@ -1109,32 +1146,39 @@ class DummyDispatcher:
         self.flushed = True
 
 class DummyStream:
-    def __init__(self, error=None):
+    def __init__(self, error=None, fileno=20):
         self.error = error
         self.closed = False
         self.flushed = False
-        self.written = ''
+        self.written = b''
+        self._fileno = fileno
     def close(self):
         if self.error:
             raise self.error
         self.closed = True
     def flush(self):
+        if self.error:
+            raise self.error
         self.flushed = True
     def write(self, msg):
         if self.error:
-            raise self.error
-        self.written +=msg
+            error = self.error
+            self.error = None
+            raise error
+        self.written += as_bytes(msg)
     def seek(self, num, whence=0):
         pass
     def tell(self):
         return len(self.written)
+    def fileno(self):
+        return self._fileno
 
 class DummyEvent:
     def __init__(self, serial='abc'):
         if serial is not None:
             self.serial = serial
 
-    def __str__(self):
+    def payload(self):
         return 'dummy event'
 
 class DummyPoller:
