@@ -7,6 +7,6 @@ counter = 0
 
 while 1:
    time.sleep(0.01)
-   print "more spewage %s" % counter
+   print("more spewage %s" % counter)
    counter += 1
    
