@@ -16,6 +16,8 @@ from supervisor.tests.base import DummyProcess
 from supervisor.tests.base import DummyProcessGroup
 from supervisor.tests.base import DummyDispatcher
 
+from supervisor.compat import StringIO
+
 try:
     import pstats
 except ImportError: # pragma: no cover
@@ -29,8 +31,7 @@ class EntryPointTests(unittest.TestCase):
         conf = os.path.join(
             os.path.abspath(os.path.dirname(__file__)), 'fixtures',
             'donothing.conf')
-        import StringIO
-        new_stdout = StringIO.StringIO()
+        new_stdout = StringIO()
         new_stdout.fileno = lambda: 1
         old_stdout = sys.stdout
         try:
@@ -44,7 +45,7 @@ class EntryPointTests(unittest.TestCase):
             sys.stdout = old_stdout
             shutil.rmtree(tempdir)
         output = new_stdout.getvalue()
-        self.assertTrue(output.find('supervisord started') != 1, output)
+        self.assertTrue('supervisord started' in output, output)
 
     if pstats:
         def test_main_profile(self):
@@ -52,8 +53,7 @@ class EntryPointTests(unittest.TestCase):
             conf = os.path.join(
                 os.path.abspath(os.path.dirname(__file__)), 'fixtures',
                 'donothing.conf')
-            import StringIO
-            new_stdout = StringIO.StringIO()
+            new_stdout = StringIO()
             new_stdout.fileno = lambda: 1
             old_stdout = sys.stdout
             try:
@@ -67,8 +67,7 @@ class EntryPointTests(unittest.TestCase):
                 sys.stdout = old_stdout
                 shutil.rmtree(tempdir)
             output = new_stdout.getvalue()
-            self.assertTrue(output.find('cumulative time, call count') != -1,
-                            output)
+            self.assertTrue('cumulative time, call count' in output, output)
 
 class SupervisordTests(unittest.TestCase):
     def tearDown(self):
@@ -149,6 +148,27 @@ class SupervisordTests(unittest.TestCase):
         supervisord.reap(once=True)
         self.assertEqual(process.finished, (1,1))
 
+    def test_reap_recursionguard(self):
+        options = DummyOptions()
+        supervisord = self._makeOne(options)
+        result = supervisord.reap(once=True, recursionguard=100)
+        self.assertEqual(result, None)
+
+    def test_reap_more_than_once(self):
+        options = DummyOptions()
+        options.waitpid_return = 1, 1
+        pconfig = DummyPConfig(options, 'process', 'process', '/bin/process1')
+        process = DummyProcess(pconfig)
+        process.drained = False
+        process.killing = True
+        process.laststop = None
+        process.waitstatus = None, None
+        options.pidhistory = {1:process}
+        supervisord = self._makeOne(options)
+
+        supervisord.reap(recursionguard=99)
+        self.assertEqual(process.finished, (1,1))
+
     def test_reap_unknown_pid(self):
         options = DummyOptions()
         options.waitpid_return = 2, 0 # pid, status
@@ -246,12 +266,15 @@ class SupervisordTests(unittest.TestCase):
         options.process_group_configs = DummyPGroupConfig(
             options, 'foo',
             pconfigs=pconfigs)
+        dummypgroup = DummyProcessGroup(options)
+        supervisord.process_groups = {None:dummypgroup}
         supervisord.handle_signal()
         self.assertEqual(supervisord.options.mood,
                          SupervisorStates.RUNNING)
         self.assertEqual(options.logs_reopened, True)
         self.assertEqual(options.logger.data[0],
                          'received SIGUSR2 indicating log reopen request')
+        self.assertEqual(dummypgroup.logs_reopened, True)
 
     def test_handle_unknown_signal(self):
         options = DummyOptions()
@@ -324,14 +347,17 @@ class SupervisordTests(unittest.TestCase):
                 'uid': None, 'stdout_logfile': None, 'stdout_capture_maxbytes': 0,
                 'stdout_events_enabled': False,
                 'stdout_logfile_backups': 0, 'stdout_logfile_maxbytes': 0,
+                'stdout_syslog': False,
                 'stderr_logfile': None, 'stderr_capture_maxbytes': 0,
                 'stderr_events_enabled': False,
                 'stderr_logfile_backups': 0, 'stderr_logfile_maxbytes': 0,
+                'stderr_syslog': False,
                 'redirect_stderr': False,
                 'stopsignal': None, 'stopwaitsecs': 10,
                 'stopasgroup': False,
                 'killasgroup': False,
-                'exitcodes': (0,2), 'environment': None, 'serverurl': None }
+                'exitcodes': (0,), 'environment': None, 'serverurl': None,
+            }
             result.update(params)
             return ProcessConfig(options, **result)
 
@@ -396,7 +422,7 @@ class SupervisordTests(unittest.TestCase):
                 'stopsignal': None, 'stopwaitsecs': 10,
                 'stopasgroup': False,
                 'killasgroup': False,
-                'exitcodes': (0,2), 'environment': None, 'serverurl': None,
+                'exitcodes': (0,), 'environment': None, 'serverurl': None,
             }
             result.update(params)
             return EventListenerConfig(options, **result)
@@ -501,7 +527,7 @@ class SupervisordTests(unittest.TestCase):
         self.assertEqual(supervisord.process_groups, {})
 
         result = supervisord.add_process_group(gconfig)
-        self.assertEqual(supervisord.process_groups.keys(), ['foo'])
+        self.assertEqual(list(supervisord.process_groups.keys()), ['foo'])
         self.assertTrue(result)
 
         group = supervisord.process_groups['foo']
@@ -545,7 +571,7 @@ class SupervisordTests(unittest.TestCase):
         supervisord.add_process_group(gconfig)
         supervisord.process_groups['foo'].unstopped_processes = [DummyProcess(None)]
         result = supervisord.remove_process_group('foo')
-        self.assertEqual(supervisord.process_groups.keys(), ['foo'])
+        self.assertEqual(list(supervisord.process_groups.keys()), ['foo'])
         self.assertTrue(not result)
 
     def test_remove_process_group_event(self):
@@ -619,7 +645,7 @@ class SupervisordTests(unittest.TestCase):
         self.assertEqual(writable.write_event_handled, True)
         self.assertEqual(error.error_handled, True)
 
-    def test_runforever_select_dispatcher_exitnow(self):
+    def test_runforever_select_dispatcher_exitnow_via_read(self):
         options = DummyOptions()
         options.poller.result = [6], []
         supervisord = self._makeOne(options)
@@ -633,6 +659,48 @@ class SupervisordTests(unittest.TestCase):
         options.test = True
         self.assertRaises(asyncore.ExitNow, supervisord.runforever)
 
+    def test_runforever_select_dispatcher_exitnow_via_write(self):
+        options = DummyOptions()
+        options.poller.result = [], [6]
+        supervisord = self._makeOne(options)
+        pconfig = DummyPConfig(options, 'foo', '/bin/foo',)
+        gconfig = DummyPGroupConfig(options, pconfigs=[pconfig])
+        pgroup = DummyProcessGroup(gconfig)
+        from supervisor.medusa import asyncore_25 as asyncore
+        exitnow = DummyDispatcher(readable=True, error=asyncore.ExitNow)
+        pgroup.dispatchers = {6:exitnow}
+        supervisord.process_groups = {'foo': pgroup}
+        options.test = True
+        self.assertRaises(asyncore.ExitNow, supervisord.runforever)
+
+    def test_runforever_select_dispatcher_handle_error_via_read(self):
+        options = DummyOptions()
+        options.poller.result = [6], []
+        supervisord = self._makeOne(options)
+        pconfig = DummyPConfig(options, 'foo', '/bin/foo',)
+        gconfig = DummyPGroupConfig(options, pconfigs=[pconfig])
+        pgroup = DummyProcessGroup(gconfig)
+        notimpl = DummyDispatcher(readable=True, error=NotImplementedError)
+        pgroup.dispatchers = {6:notimpl}
+        supervisord.process_groups = {'foo': pgroup}
+        options.test = True
+        supervisord.runforever()
+        self.assertEqual(notimpl.error_handled, True)
+
+    def test_runforever_select_dispatcher_handle_error_via_write(self):
+        options = DummyOptions()
+        options.poller.result = [], [6]
+        supervisord = self._makeOne(options)
+        pconfig = DummyPConfig(options, 'foo', '/bin/foo',)
+        gconfig = DummyPGroupConfig(options, pconfigs=[pconfig])
+        pgroup = DummyProcessGroup(gconfig)
+        notimpl = DummyDispatcher(readable=True, error=NotImplementedError)
+        pgroup.dispatchers = {6:notimpl}
+        supervisord.process_groups = {'foo': pgroup}
+        options.test = True
+        supervisord.runforever()
+        self.assertEqual(notimpl.error_handled, True)
+
     def test_runforever_stopping_emits_events(self):
         options = DummyOptions()
         supervisord = self._makeOne(options)
