@@ -1,18 +1,22 @@
 import os
 import re
-import cgi
 import time
 import traceback
-import urllib
 import datetime
 
+from supervisor import templating
+
+from supervisor.compat import urllib
+from supervisor.compat import urlparse
+from supervisor.compat import as_string
+from supervisor.compat import PY2
+from supervisor.compat import unicode
+
 from supervisor.medusa import producers
 from supervisor.medusa.http_server import http_date
 from supervisor.medusa.http_server import get_header
 from supervisor.medusa.xmlrpc_handler import collector
 
-import meld3
-
 from supervisor.process import ProcessStates
 from supervisor.http import NOT_DONE_YET
 
@@ -62,7 +66,7 @@ class DeferredWebProducer:
         for header in headers:
             self.request[header] = headers[header]
 
-        if not self.request.has_key('Content-Type'):
+        if 'Content-Type' not in self.request:
             self.request['Content-Type'] = 'text/plain'
 
         if headers.get('Location'):
@@ -91,8 +95,8 @@ class DeferredWebProducer:
         elif self.request.version == '1.1':
             if connection == 'close':
                 close_it = 1
-            elif not self.request.has_key ('Content-Length'):
-                if self.request.has_key ('Transfer-Encoding'):
+            elif 'Content-Length' not in self.request:
+                if 'Transfer-Encoding' in self.request:
                     if not self.request['Transfer-Encoding'] == 'chunked':
                         close_it = 1
                 elif self.request.use_chunked:
@@ -118,6 +122,12 @@ class DeferredWebProducer:
                 [outgoing_header, outgoing_producer]
                 )
         else:
+            # fix AttributeError: 'unicode' object has no attribute 'more'
+            if PY2 and (len(self.request.outgoing) > 0):
+                body = self.request.outgoing[0]
+                if isinstance(body, unicode):
+                    self.request.outgoing[0] = producers.simple_producer (body)
+
             # prepend the header
             self.request.outgoing.insert(0, outgoing_header)
             outgoing_producer = producers.composite_producer (
@@ -146,7 +156,7 @@ class ViewContext:
 
 class MeldView:
 
-    content_type = 'text/html'
+    content_type = 'text/html;charset=utf-8'
     delay = .5
 
     def __init__(self, context):
@@ -155,7 +165,7 @@ class MeldView:
         if not os.path.isabs(template):
             here = os.path.abspath(os.path.dirname(__file__))
             template = os.path.join(here, template)
-        self.root = meld3.parse_xml(template)
+        self.root = templating.parse_xml(template)
         self.callback = None
 
     def __call__(self):
@@ -169,9 +179,12 @@ class MeldView:
         headers['Pragma'] = 'no-cache'
         headers['Cache-Control'] = 'no-cache'
         headers['Expires'] = http_date.build_http_date(0)
-        response['body'] = body
+        response['body'] = as_string(body)
         return response
 
+    def render(self):
+        pass
+
     def clone(self):
         return self.root.clone()
 
@@ -187,11 +200,7 @@ class TailView(MeldView):
             processname = form['processname']
             offset = 0
             limit = form.get('limit', '1024')
-            if limit.isdigit():
-                limit = min(-1024, int(limit) * -1)
-            else:
-                limit = -1024
-
+            limit = min(-1024, int(limit)*-1 if limit.isdigit() else -1024)
             if not processname:
                 tail = 'No process name found'
             else:
@@ -199,7 +208,7 @@ class TailView(MeldView):
                 try:
                     tail = rpcinterface.readProcessStdoutLog(processname,
                                                              limit, offset)
-                except RPCError, e:
+                except RPCError as e:
                     if e.code == Faults.NO_FILE:
                         tail = 'No file for %s' % processname
                     else:
@@ -223,7 +232,7 @@ class TailView(MeldView):
         else:
             refresh_anchor.deparent()
 
-        return root.write_xhtmlstring()
+        return as_string(root.write_xhtmlstring())
 
 class StatusView(MeldView):
     def actions_for_process(self, process):
@@ -330,7 +339,7 @@ class StatusView(MeldView):
                         bool_or_callback = (
                             rpcinterface.supervisor.startProcess(namespec)
                             )
-                    except RPCError, e:
+                    except RPCError as e:
                         if e.code == Faults.NO_FILE:
                             msg = 'no such file'
                         elif e.code == Faults.NOT_EXECUTABLE:
@@ -353,7 +362,7 @@ class StatusView(MeldView):
                         def startprocess():
                             try:
                                 result = bool_or_callback()
-                            except RPCError, e:
+                            except RPCError as e:
                                 if e.code == Faults.SPAWN_ERROR:
                                     msg = 'spawn error'
                                 elif e.code == Faults.ABNORMAL_TERMINATION:
@@ -379,10 +388,10 @@ class StatusView(MeldView):
                         bool_or_callback = (
                             rpcinterface.supervisor.stopProcess(namespec)
                             )
-                    except RPCError, e:
+                    except RPCError as e:
+                        msg = 'unexpected rpc fault [%d] %s' % (e.code, e.text)
                         def stoperr():
-                            return 'unexpected rpc fault [%d] %s' % (
-                                e.code, e.text)
+                            return msg
                         stoperr.delay = 0.05
                         return stoperr
 
@@ -390,7 +399,7 @@ class StatusView(MeldView):
                         def stopprocess():
                             try:
                                 result = bool_or_callback()
-                            except RPCError, e:
+                            except RPCError as e:
                                 return 'unexpected rpc fault [%d] %s' % (
                                     e.code, e.text)
                             if result is NOT_DONE_YET:
@@ -431,10 +440,10 @@ class StatusView(MeldView):
                     try:
                         callback = rpcinterface.supervisor.clearProcessLogs(
                             namespec)
-                    except RPCError, e:
+                    except RPCError as e:
+                        msg = 'unexpected rpc fault [%d] %s' % (e.code, e.text)
                         def clearerr():
-                            return 'unexpected rpc fault [%d] %s' % (
-                                e.code, e.text)
+                            return msg
                         clearerr.delay = 0.05
                         return clearerr
 
@@ -463,7 +472,7 @@ class StatusView(MeldView):
                     return NOT_DONE_YET
                 if message is not None:
                     server_url = form['SERVER_URL']
-                    location = server_url + '?message=%s' % urllib.quote(
+                    location = server_url + "/" + '?message=%s' % urllib.quote(
                         message)
                     response['headers']['Location'] = location
 
@@ -474,10 +483,8 @@ class StatusView(MeldView):
             )
 
         processnames = []
-        groups = supervisord.process_groups.values()
-        for group in groups:
-            gprocnames = group.processes.keys()
-            for gprocname in gprocnames:
+        for group in supervisord.process_groups.values():
+            for gprocname in group.processes.keys():
                 processnames.append((group.config.name, gprocname))
 
         processnames.sort()
@@ -547,7 +554,7 @@ class StatusView(MeldView):
         copyright_year = str(datetime.date.today().year)
         root.findmeld('copyright_date').content(copyright_year)
 
-        return root.write_xhtmlstring()
+        return as_string(root.write_xhtmlstring())
 
 class OKView:
     delay = 0
@@ -605,14 +612,14 @@ class supervisor_ui_handler:
         form = {}
         cgi_env = request.cgi_environment()
         form.update(cgi_env)
-        if not form.has_key('QUERY_STRING'):
+        if 'QUERY_STRING' not in form:
             form['QUERY_STRING'] = ''
 
         query = form['QUERY_STRING']
 
         # we only handle x-www-form-urlencoded values from POSTs
-        form_urlencoded = cgi.parse_qsl(data)
-        query_data = cgi.parse_qs(query)
+        form_urlencoded = urlparse.parse_qsl(data)
+        query_data = urlparse.parse_qs(query)
 
         for k, v in query_data.items():
             # ignore dupes
@@ -636,8 +643,7 @@ class supervisor_ui_handler:
             # this should never happen if our match method works
             return
 
-        response = {}
-        response['headers'] = {}
+        response = {'headers': {}}
 
         viewclass = viewinfo['view']
         viewtemplate = viewinfo['template']
